/* test_direct.cpp - Demonstrate usage of miarn code (without any
 *                   supporting classes) to extract features from
 *                   laser range data and track people (i.e., pairs of
 *                   legs).
 *
 * Scott Livingston
 * 18 March 2009
 */

#include <iostream>
#include <string>
using namespace std;

#include <cstdlib>
#include <cstring>
#include <cmath>

#include <libplayerc++/playerc++.h>
using namespace PlayerCc;

#include "feature_geometry.hpp"
#include "feature_person.hpp"


int main( int argc, char *argv[] )
{
	LaserFeatureX laserFeature;
	FeatureLegTracker legTrack;

	uint numCircles, numLines, numPLeg, numP2Legs;

	player_point_2d_t pp;
	bool check2legs; // This name comes from Joao's pv3d code.

	string hostname( "localhost" );
	int port = 6665;

	if (argc > 3 || (argc == 2 && !strcmp( argv[1], "-h" ))) {
		cerr << "Usage: " << argv[0] << " [hostname] [port]" << endl;
		return 1;
	} else if (argc == 2) { // Only hostname specified
		hostname = argv[1];
	} else if (argc == 3) { // Hostname and port number both specified
		hostname = argv[1];
		port = atof( argv[2] );
	}
	
	try {

		PlayerClient robot( hostname.c_str(), port );
		LaserProxy laserp( &robot, 0 );

		// Main loop
		while (true) {

			robot.Read();

			if (!laserp.IsValid() || !laserp.IsFresh())
				continue;

			laserFeature.fdata.clear();
			laserFeature.ranges.clear();
			laserFeature.point_xy.clear();

			for (uint i = 0; i < laserp.GetCount(); i++) {
				laserFeature.ranges.push_back( laserp.GetRange(i) );
				pp = laserp.GetPoint(i);
				laserFeature.point_xy.push_back( xy( pp.px, pp.py ) );
			}

			laserFeature.Segmentation();

			for (uint i = 0; i < laserFeature.segments.size(); i++) {

				// Fit lines
				laserFeature.RecursiveLineFitting( laserFeature.segments[i].begin, laserFeature.segments[i].end );

				// Fit circles
				check2legs = !laserFeature.FitArc( laserFeature.segments[i].begin, laserFeature.segments[i].end );

				// Find legs
				laserFeature.FindLeg( laserFeature.segments[i].begin, laserFeature.segments[i].end, check2legs );

			}

			// Leg tracking!
			legTrack.fdata.clear();
			legTrack.fdata = laserFeature.fdata;
			legTrack.LCReset();
			legTrack.CreateIncoming();
			legTrack.LegMatchInTime();
			legTrack.LegClean();
			legTrack.LegCreate();
			legTrack.PersonUpdate();
			legTrack.PersonCreate();
			legTrack.FillFeature();

			// Print features found
			numCircles = numLines = numPLeg = numP2Legs = 0;
			for (uint i = 0; i < laserFeature.fdata.size(); i++) {
				switch (laserFeature.fdata[i].type) {
				case MIARN_FEATURE_TYPE_ARC:
					numCircles++;
					break;
				case MIARN_FEATURE_TYPE_LINE:
					numLines++;
					break;
				case MIARN_FEATURE_TYPE_PLEG:
					numPLeg++;
					break;
				case MIARN_FEATURE_TYPE_2LEGS:
					numP2Legs++;
					break;
				case MIARN_FEATURE_TYPE_SEGMENT:
					// Ignore these.
					break;
				default:
					cerr << "Error: unrecognized feature type code: " << laserFeature.fdata[i].type << endl;
				}
			}
			if (numCircles > 0 || numLines > 0 || numPLeg > 0 || numP2Legs > 0)
				cout << "Found on this cycle"
					 << "\n  circles: " << numCircles
					 << "\n  lines: " << numLines
					 << "\n  pleg: " << numPLeg
					 << "\n  p2legs: " << numP2Legs << endl;

			// List extracted "people" (actually, pairs of legs)
			for (uint i = 0; i < legTrack.fdata_out.size(); i++) {
				switch (legTrack.fdata_out[i].type) {
				case MIARN_FEATURE_TYPE_LEG:
					// Ignore for now...
					break;
				case MIARN_FEATURE_TYPE_PERSON:
					cout << "Person found at ( " << legTrack.fdata_out[i].pos[0] << ", "
						 << legTrack.fdata_out[i].pos[1] << " ), at angle "
						 << atan2f( legTrack.fdata_out[i].pos[1], legTrack.fdata_out[i].pos[0] )*180/3.14
						 << " deg" << endl;
					break;
				default:
					cerr << "Error: unrecognized feature type code: " << laserFeature.fdata[i].type << endl;
				}
			}

		}

	} catch (PlayerError e) {
		cerr << e << endl;
		return -1;
	}
	
	return 0;
}
