/* PersonTracker.hpp - Track a person using a combination of IAV
 *                     laser-based and several vision based trackers.
 *
 * Scott Livingston <SCL@utk.edu>
 * April-May 2009
 *
 * NOTES: - Though laser and camera proxies are internally managed,
 *          the main Player robot sensor stream must be updated
 *          externally, i.e., you should call something like
 *          robot.Read() before calling the Update method of this
 *          class.
 *
 *        - Unless otherwise specified, any function with bool return
 *          type returns "true" on success, "false" on failure.
 *
 *        - Distance units are in meters, speed units are in meters/second.
 *
 *        - Storing target position in robot coordinate frame will
 *          impose a heavy computational burden when the robot moves
 *          (and thus ALL candidate target positions must be
 *          transformed).
 *
 *        - Times are reported (and stored) in seconds (microsecond resolution) since the Epoch.
 *
 *        - Note that radius to target person (called `rho') is NOT
 *          updated when only vision-based trackers can find her. In
 *          general, updates based only on vision trackers are poor
 *          and uncertainty in position should increase with time.
 *          Such updating behavior is not in the current
 *          implementation, however.
 *
 *        - Uncertainty from features extracted from laser-based (IAV)
 *          leg detection is assumed to be directly proportional to
 *          the square of the distance to the `leg' feature.
 *
 *        - Note that it may be worth considering storing a pointer to
 *          the robot proxy (a PlayerClient object) so that fresh
 *          sensor data (via Read method) can be acquired internally
 *          by PersonTracker object rather than relying on the user to
 *          do so (before calling the Update method of PersonTracker).
 *
 */

#ifndef _PERSONTRACKER_HPP_
#define _PERSONTRACKER_HPP_


#include <time.h>

#include <cv.h>
#include <libplayerc++/playerc++.h>


#include "ColorHistTracker.hpp"
#include "KernelObjTracker.hpp"
#include "IAVLaserFeatureProxy.hpp"


// Tracker states
#define PERSONTRACK_UNINIT 0 // uninitialized (i.e., no idea where target person is)
#define PERSONTRACK_KNOWN  1 // known (... or at least, we'd like to think so)


typedef unsigned int uint;


typedef struct {
	double rho, theta;
	double uncertainty; // `variance' for purposes of Kalman filtering.
} target_position_t;


class PersonTracker {

public:
	PersonTracker( PlayerCc::PlayerClient *aPc_laser, uint32_t aIndex_laser,
				   PlayerCc::PlayerClient *aPc_camera, uint32_t aIndex_camera,
				   const IplImage *initImg, CvRect initRegion,
				   int numBins = 16 ); // For initializing both ColorHistTracker and KernelObjTracker
	~PersonTracker();

	/* (Re)initialize target (training) image;
	   note that this method does NOT alter nor clear the current
	   estimated target person position. */
	bool InitTarget( const IplImage *initImg, CvRect initRegion,
					 int numBins = 16 ); // For initializing both ColorHistTracker and KernelObjTracker

	/* Attempt to find target person (if uninitialized) or update
	   position estimate given current data.
       Returns tracker state (see definitions above) on success,
               -1 on error. */
	int Update();

	// Clear estimated (tracked) person position and set tracker state to `uninitialized'.
	bool ResetTrackedPos();

	/* Obtain current estimated position of target person.
	   Note that this is meaningless if tracker state is unitialized. */
	target_position_t GetEstimate() const;

	// (See definitions of state above.)
	unsigned char GetTrackerState() const;

	/* In robot coordinate frame, calibrate camera assuming the object
	   currently tracked by vision is at given (x,y) coordinate.
	   maxStabilizationCount is the maximum number of vision-based
	   tracker updates while waitingi for the target candidate
	   position to stabilize */
	bool CalibrateCam( PlayerCc::PlayerClient *robot, // robot to which the camera is attached
					   double x, double y,
					   unsigned int maxStabilizationCount = 10 );

	/* Returns timestamp (seconds since Epoch) since last tracker update.
	   Note that this is meaningless if tracker state is unitialized. */
	double GetLastUpdateTime() const;

	// Returns seconds (microsecond resolution) since last target position update.
	// (Returns -1.0 on error.)
	double GetTimeSinceUpdate() const;

	float GetMaxPersonRadius() const { return maxPersonRadius; }
	float GetMaxPersonSpeed() const { return maxPersonSpeed; }
	float GetPixels2RadRatio() const { return pixel2rad; }
	float GetMaxVisionLaserMismatch() const { return maxVisionLaserMismatch; }

	// These set functions include basic error-checking (e.g., ensure non-negativity).
	bool SetMaxPersonRadius( float newRadius );
	bool SetMaxPersonSpeed( float newSpeed );
	bool SetPixels2RadRatio( float newRatio );
	bool SetMaxVisionLaserMismatch( float newMaxMismatch );

	// For approximating transformations to robot coordinate frame between tracking updates.
	bool SetRobotVelocity( float newTransVelocity, // Translational velocity
						   float newRotVelocity ); // Rotational velocity

	// Some manipulation routines for the vision-based trackers
	bool SetVisionInitCenter( int x0, int y0 ); // Set center of search field (so-called `basin of operation').
	bool SetVisionBandwidth( double newBW ); // Relevant for kernel-based (vision) object tracker
	double GetVisionBandwidth() const;
	int GetVisionCenterX() const;
	int GetVisionCenterY() const;
	CvRect GetTargetRegion() const;

	/* Since running more than one laser proxy in the same program
	   makes Player 2.0.3 disgruntled, here are a few routines for
	   accessing the laser proxy internal to PersonTracker. */
	uint32_t GetCount() const { return iavLaserProxy.GetCount(); }
	double GetRange( uint32_t aIndex ) const { return iavLaserProxy.GetRange( aIndex ); }
	double GetBearing( uint32_t aIndex ) const { return iavLaserProxy.GetBearing( aIndex ); }
	double MinLeft() { return iavLaserProxy.MinLeft(); }
	double MinRight() { return iavLaserProxy.MinRight(); }
	double operator[] (uint32_t index) const { return GetRange( index ); }
	//double GetMaxRange() const { return iavLaserProxy.GetMaxRange(); }

private:
	// Tracked person data
	unsigned char trackedState; // State of the tracked object (see DEFINEs above).
	target_position_t estimatedPos; // Estimated position of tracked person in robot coordinate frame
	timeval timestamp; // Time of last update (only relevant if tracked_state is PERSONTRACK_KNOWN).

	PlayerCc::CameraProxy cameraProxy;

	IAVLaserFeatureProxy iavLaserProxy;
	ColorHistTracker cht;
	KernelObjTracker kot;

	/* Used for approximating change in robot's coordinate frame with
	   respect to the inertial coordinate frame. Thus the estimate of
	   the target person's position can be appropriately adjusted
	   before attempting an update. */
	float robotTransVelocity, // Translational velocity
		robotRotVelocity;     // Rotational velocity

	float maxPersonRadius; // Maximum distance between legs for a single person
	float maxPersonSpeed;  // Maximum person speed
	float pixel2rad;       // Radians per pixel

	/* Maximum difference between theta (i.e., angle part of position)
	   estimated by vision-based tracking and laser-based
	   tracking. Results depend on what state the person tracker is
	   in...

       Uninitialized                 - Differences in theta below this threshold
                                       allow initialization and thus changing to
                                       `known' tracker state.

       Known (i.e., active tracking) - Differences in theta above this threshold
                                       cause person tracker to update based only
                                       on one of the two sensor modalities, whichever
                                       is closest to previous estimated target position.
	*/
	float maxVisionLaserMismatch; // in radians.

	IplImage *prevImg; /* Used internally for detecting motion;
						  updated after a (successful) call to Update
						  method (for position tracking). */

	// Internal methods:
	bool UpdateVision(); /* Attempt to update vision-based component
							using current camera data. */

};


#endif
