/* testPT.cpp - Test the PersonTracker class.
 *
 * Scott Livingston <SCL@utk.edu>
 * Originally created: 27 April 2009
 *
 */


#include <opencv/cv.h>
#include <opencv/highgui.h>

#include <libplayerc++/playerc++.h>
using namespace PlayerCc;

#include <cmath>
#include <ctime>
#include <cstdio>
#include <iostream>
#include <fstream>
using namespace std;

#include "PersonTracker.hpp"


#define MAX_SPEED .2
#define MIN_DIST .5


// obstacle avoidance
bool oaStop( PersonTracker *ptrack,
		double minAngle, double maxAngle, double minDist )
{
	for (unsigned int i = 0; i < ptrack->GetCount() && ptrack->GetBearing(i) <= maxAngle; i++) {
		if (ptrack->GetBearing(i) < minAngle)
			continue;

		if (ptrack->GetRange(i) < minDist)
			return true;
	}
	return false;
}


// Returns `true' if given (x,y) lies within circle centered at (x0,y0) with specified radius.
bool visitBall( double x0, double y0, double radius, double x, double y )
{
	if (((x0-x)*(x0-x) + (y0-y)*(y0-y)) <= radius*radius)
		return true;
	else
		return false;
}

// Regarding experimental recording from April 19,
// Return numbered point in which given (x,y) position lies, or -1 if none.
int getPointLabel( double x, double y, double maxDist )
{
	if (visitBall( 1.524, 1.524, maxDist, x, y )) { // Point a, 1
		return 1;
	} else if (visitBall( 1.524, -1.524, maxDist, x, y )) { // Point b, 2
		return 2;
	} else if (visitBall( 2.743, 1.524, maxDist, x, y )) { // Point c, 3
		return 3;
	} else if (visitBall( 2.743, -0.305, maxDist, x, y )) { // Point d, 4
		return 4;
	} else if (visitBall( 2.743, -1.524, maxDist, x, y )) { // Point e, 5
		return 5;
	} else if (visitBall( 4.267, 1.524, maxDist, x, y )) { // Point f, 6
		return 6;
	} else if (visitBall( 4.267, -1.219, maxDist, x, y )) { // Point g, 7
		return 7;
	} else {
		return -1;
	}
}


int main( int argc, char **argv )
{
	IplImage *img;
	int updateResult; // Returned by Update method of PersonTracker class.
	target_position_t targetPos;
	bool printedUnknownStatus = false;
	int visitPoint;
	int lastVisitPoint;
	timeval start_time, current_time;
	float runtime, elapsed_time;

	if (argc != 9 && argc != 8) {
		cerr << "Usage: " << argv[0] << " init_image x0 y0 width height bandwidth filename [runtime]" << endl;
		return 1;
	}

	ofstream datfile( argv[7] );
	if (datfile.fail()) {
		cerr << "Error: failed to open " << argv[7] << " for writing." << endl;
		return -1;
	}

	if (argc == 9) {
		runtime = atof( argv[8] );
		if (runtime < 0) {
			cerr << "Error: runtime must be non-negative." << endl;
			return -1;
		}
	} else {
		runtime = -1;
	}

	// Attempt to load target initialization image
	if ((img = cvLoadImage( argv[1] )) == NULL) {
		cerr << "Error: Could not open image file: " << argv[1] << endl;
		return -1;
	}

	// Verify sign of given region arguments
	if (atoi(argv[2]) < 0 || atoi(argv[3]) < 0
		|| atoi(argv[4]) < 0 || atoi(argv[5]) < 0 ) {
		cerr << "Error: Given region specs (x0, y0, width, height) must be non-negative." << endl;
		return -1;
	}

	// ..and sign of desired vision-tracking bandwidth
	if (atof( argv[6] ) < 0) {
		cerr << "Error: Bandwidth must be non-negative." << endl;
		return -1;
	}

	PlayerClient robot( "localhost" );
	robot.SetDataMode( PLAYER_DATAMODE_PUSH );
	robot.SetReplaceRule( true );
	Position2dProxy pp( &robot, 0 );

	// Attempt to instantiate and initialize a ColorHistTracker object
	PersonTracker ptrack( &robot, 0, &robot, 0,
						  img, cvRect( atoi(argv[2]), atoi(argv[3]),
									   atoi(argv[4]), atoi(argv[5]) ) );

	if (!ptrack.SetVisionBandwidth( atof( argv[6] ) )) {
		cerr << "Error: Failed to initialize with given bandwidth: " << atof( argv[6] ) << endl;
		return -1;
	}

	cvReleaseImage( &img ); // Init target image is no longer needed.
	img = NULL;

	

	try {

		// Before entering loop, attempt to calibrate camera
		//ptrack.CalibrateCam( &robot, 2., 1., 100 );

		// Init timer
		if (gettimeofday( &start_time, NULL ) == -1) {
			perror( "gettimeofday" );
			return -1;
		}

		while (true) { // Track person ad infinitum.

			// Get most recent sensor data from robot server.
			robot.Read();

			if (gettimeofday( &current_time, NULL ) == -1) {
				perror( "gettimeofday" );
				return -1;
			}

			// Update person tracker.
			updateResult = ptrack.Update();

			// Update timer.
			elapsed_time = current_time.tv_sec + current_time.tv_usec/(float)(1e6)
				- (start_time.tv_sec + start_time.tv_usec/(float)(1e6));

			if (runtime >= 0. && elapsed_time > runtime)
				break;

			switch (updateResult) {

			case -1:
				cerr << "PersonTracker::Update returned error." << endl;
				break;

			case PERSONTRACK_UNINIT:
				if (!printedUnknownStatus) {
					cout << "Target position is unknown." << endl;
					printedUnknownStatus = true;
					//pp.SetSpeed( 0, 0 );
					//ptrack.SetRobotVelocity( 0, 0 );
				}
				if (targetPos.theta > 0) {
					pp.SetSpeed( 0, .1 );
				} else {
					pp.SetSpeed( 0, -.1 );
				}
				break;

			case PERSONTRACK_KNOWN:
				targetPos = ptrack.GetEstimate();
				cout << "Esimated target position is ( " // in (x,y) coordinates
					 << targetPos.rho*cos( targetPos.theta ) << ", "
					 << targetPos.rho*sin( targetPos.theta ) << " )" << endl;

				if (abs(targetPos.theta) > 0.17) {
					pp.SetSpeed( 0, targetPos.theta/3 );
					//ptrack.SetRobotVelocity( 0, targetPos.theta/3 );
				} else if (targetPos.rho > .3 && !oaStop( &ptrack, -M_PI/8, M_PI/8, .3 )) {
					pp.SetSpeed( targetPos.rho/10 > MAX_SPEED ? MAX_SPEED : targetPos.rho/10, 0 );
					//ptrack.SetRobotVelocity( targetPos.rho/10 > MAX_SPEED ? MAX_SPEED : targetPos.rho/10, 0 );
				} else {
					pp.SetSpeed( 0, 0 );
					//ptrack.SetRobotVelocity( 0, 0 );
				}
				//}
				//visitPoint = getPointLabel( targetPos.rho*cos( targetPos.theta ), targetPos.rho*sin( targetPos.theta ), .5 );

				// Print position to trajectory log file (timestamped).
				//datfile << elapsed_time << " "
				//		<< targetPos.rho*cos( targetPos.theta ) << " "
				//		<< targetPos.rho*sin( targetPos.theta ) << endl;

				if (visitPoint > 0 && (lastVisitPoint != visitPoint || printedUnknownStatus == true)) {
					//cout << "Near point " << char( visitPoint+96 ) << endl;
					lastVisitPoint = visitPoint;
				}

				printedUnknownStatus = false;
				break;

			default:
				cerr << "PersonTracker::Update returned unrecognized code:" << updateResult << endl;
				break;
			}
			
		}
	} catch( PlayerError e) {
		cerr << e << endl;
		return -1;
	}

	datfile.close();

	return 0;
}
