/* testCHT_player.cpp - Example using ColorHistTracker class that
 *                      reads images from camera device a la Player.
 *
 * Scott Livingston <SCL@utk.edu>
 * Originally created: 1 April 2009
 *
 */


#include <cv.h>
#include <highgui.h>

#include <libplayerc++/playerc++.h>
using namespace PlayerCc;

#include <cstdlib>
#include <iostream>
using namespace std;

#include "ColorHistTracker.hpp"


typedef unsigned int uint;


int main( int argc, char **argv )
{
	IplImage *img, *prevImg,
		*dispImage; // Use bounding boxes for visual feedback

	uint cam_width, cam_height, cam_depth;
	uint8_t *imgBuf;

	uint i, j; // generic counters

	if (argc != 6) {
		cerr << "Usage: " << argv[0] << " init_image x0 y0 width height" << endl;
		return 1;
	}

	PlayerClient robot( "localhost" );
	CameraProxy camp( &robot, 0 );

	// Determine camera image specs
	robot.Read();
	cam_width = camp.GetWidth();
	cam_height = camp.GetHeight();
	cam_depth = camp.GetDepth();

	// Attempt to load target initialization image
	if ((img = cvLoadImage( argv[1] )) == NULL) {
		cerr << "Error: Could not open image file: " << argv[1] << endl;
		return -1;
	}

	// Verify sign of given region arguments
	if (atoi(argv[2]) < 0 || atoi(argv[3]) < 0
		|| atoi(argv[4]) < 0 || atoi(argv[5]) < 0 ) {
		cerr << "Error: Given region specs (x0, y0, width, height) must be non-negative." << endl;
		return -1;
	}

	// Attempt to instantiate and initialize a ColorHistTracker object
	ColorHistTracker cht( img, cvRect( atoi(argv[2]), atoi(argv[3]),
									   atoi(argv[4]), atoi(argv[5]) ),
						  16,     //number of histogram bins
						  .4,     //threshold
						  1024 ); //normalization factor
	cvReleaseImage( &img ); // Init target image is no longer needed.
	img = NULL;

	// Image buffer and IplImage structure
	imgBuf = new uint8_t[cam_width*cam_height*cam_depth];
	img = cvCreateImage( cvSize(cam_width, cam_height), IPL_DEPTH_8U, 3 );

	// Open a window for visual feedback and enter main (tracking) loop
	cvNamedWindow( "track results", 1 );
	//cvNamedWindow( "prep", 1 );
	prevImg = NULL; // Indicate no previous image yet
	dispImage = NULL;
	
	try {
		while (cvWaitKey(10) != 27) {

			robot.Read();
			camp.Decompress();
			camp.GetImage( imgBuf );

			// Copy image from camera proxy
			for (i = 0; i < cam_width; i++) {
				for (j = 0; j < cam_height; j++) {
					img->imageData[cam_width * j*3 + i*3 + 0] = (char)imgBuf[cam_width * j*3 + i*3 + 2];
					img->imageData[cam_width * j*3 + i*3 + 1] = (char)imgBuf[cam_width * j*3 + i*3 + 1];
					img->imageData[cam_width * j*3 + i*3 + 2] = (char)imgBuf[cam_width * j*3 + i*3 + 0];
				}
			}

			// Handle special case of first loop cycle
			if (prevImg == NULL) {
				prevImg = cvCloneImage( img );
				continue;
			}

			// Color histogram-based tracking
			cht.GetMotionRegions( prevImg, img );
			//cht.MaxNCandidateRegions( 7 ); // Area-based pruning
			//cht.CompareRegions( img );
			cht.CompareDerivRegions( img );
			cht.OrderByArea(); //Area-based sorting of "target regions"
			cht.PruneByArea( 1000 );
			cht.OrderByCorrelation(); //Correlation magnitude-based sorting of "target regions"

			// Draw rectangles on image to show found "target regions"
			if (dispImage)
				cvReleaseImage( &dispImage );
			dispImage = cvCloneImage( img );
			if (cht.GetCount() > 0)
				cout << "Correlations:\n";
			for (uint i = 0; i < cht.GetCount() && i < 2; i++) {
				cvRectangle( dispImage,
							 cvPoint(cht[i].rect.x, cht[i].rect.y),
							 cvPoint(cht[i].rect.x+cht[i].rect.width, cht[i].rect.y+cht[i].rect.height),
							 CV_RGB( 255*cht[i].r, 0, 0 ), 5 );
				cout << cht[i].r << ' ';
			}
			if (cht.GetCount() > 0)
				cout << endl;
			cvShowImage( "track results", dispImage );
		
			// For debugging, freeze frame when at least one "target region" is found
			if (cht.GetCount() > 0)
			  cvWaitKey(0);

			// Prepare for next loop cycle
			cvCopy( img, prevImg );
		}
	} catch (PlayerError e) {
		cerr << e << endl;
		return -1;
	}

	cvDestroyWindow( "track results" );
	//cvDestroyWindow( "prep" );

	// Release image buffer and structures
	if (img != NULL)
		cvReleaseImage( &img );
	if (prevImg != NULL)
		cvReleaseImage( &prevImg );
	delete [] imgBuf;

	return 0;
}
